import functools
import sys
import types
import typing as ta


_CLS_DCT_ATTR_SETS = [
    {
        '__module__',
        '__qualname__',
    },
    {
        '__all__',
    },
]


def _skip_cls_dct_frames(f: types.FrameType) -> types.FrameType:
    if sys.implementation.name == 'pypy':
        if f.f_code is functools.partial.__call__.__code__:  # noqa
            return _skip_cls_dct_frames(f.f_back)  # type: ignore

    return f


def is_possibly_cls_dct(dct: ta.Mapping[str, ta.Any]) -> bool:
    return any(all(a in dct for a in s) for s in _CLS_DCT_ATTR_SETS)


def get_caller_cls_dct(offset: int = 0) -> ta.MutableMapping[str, ta.Any]:
    f = sys._getframe(offset + 2)  # noqa
    cls_dct = _skip_cls_dct_frames(f).f_locals
    if not is_possibly_cls_dct(cls_dct):
        raise TypeError(cls_dct)
    return cls_dct


class ClassDctFn:
    def __init__(self, fn: ta.Callable, offset: int | None = None, *, wrap=True) -> None:
        super().__init__()

        self._fn = fn
        self._offset = offset if offset is not None else 1

        if wrap:
            functools.update_wrapper(self, fn)

    def __get__(self, instance, owner=None):
        return type(self)(self._fn.__get__(instance, owner), self._offset)  # noqa

    def __call__(self, *args, **kwargs):
        try:
            cls_dct = kwargs.pop('cls_dct')
        except KeyError:
            f = sys._getframe(self._offset)  # noqa
            cls_dct = _skip_cls_dct_frames(f).f_locals

        if not is_possibly_cls_dct(cls_dct):
            raise TypeError(cls_dct)

        return self._fn(cls_dct, *args, **kwargs)


def cls_dct_fn(offset=1, *, wrap=True):  # noqa
    def outer(fn):
        return ClassDctFn(fn, offset, wrap=wrap)

    return outer



import itertools
import typing as ta


T = ta.TypeVar('T')


def mut_unify_sets(sets: ta.Iterable[set[T]]) -> list[set[T]]:
    rem: list[set[T]] = list(sets)
    ret: list[set[T]] = []
    while rem:
        cur = rem.pop()
        while True:
            moved = False
            for i in range(len(rem) - 1, -1, -1):
                if any(e in cur for e in rem[i]):
                    cur.update(rem.pop(i))
                    moved = True
            if not moved:
                break
        ret.append(cur)
    if ret:
        all_ = set(itertools.chain.from_iterable(ret))
        num = sum(map(len, ret))
        if len(all_) != num:
            raise ValueError('Length mismatch')
    return ret


def unify_sets(sets: ta.Iterable[ta.AbstractSet[T]]) -> list[set[T]]:
    return mut_unify_sets([set(s) for s in sets])



"""
TODO:
 - more cache-recursive reuse - fields, mro, etc
"""
import dataclasses as dc
import sys
import typing as ta
import weakref

from ... import cached
from ... import check
from ... import collections as col
from ... import lang
from ... import reflect as rfl
from .fields import field_type
from .internals import FIELDS_ATTR
from .internals import FieldType
from .internals import Params
from .internals import is_kw_only
from .metadata import METADATA_ATTR
from .metadata import Metadata
from .metadata import get_merged_metadata
from .params import PARAMS_ATTR
from .params import ParamsExtras
from .params import get_params
from .params import get_params_extras
from .utils import Namespace


MISSING = dc.MISSING


##


def get_cls_annotations(cls: type) -> ta.Mapping[str, ta.Any]:
    # Does not use ta.get_type_hints because that's what std dataclasses do [1]. Might be worth revisiting? A part
    # of why they don't is to not import typing for efficiency but we don't care about that degree of startup speed.
    # [1]: https://github.com/python/cpython/blob/54c63a32d06cb5f07a66245c375eac7d7efb964a/Lib/dataclasses.py#L985-L986  # noqa
    return rfl.get_annotations(cls)


##


class ClassInfo:
    def __init__(self, cls: type, *, _constructing: bool = False) -> None:
        check.isinstance(cls, type)
        self._constructing = _constructing
        if not _constructing:
            check.arg(dc.is_dataclass(cls))
        super().__init__()
        self._cls: type = cls

    @property
    def cls(self) -> type:
        return self._cls  # noqa

    @cached.property
    def globals(self) -> Namespace:
        if self._cls.__module__ in sys.modules:
            return sys.modules[self._cls.__module__].__dict__
        else:
            return {}

    @cached.property
    def cls_annotations(self) -> ta.Mapping[str, ta.Any]:
        return get_cls_annotations(self._cls)

    ##

    @cached.property
    def params(self) -> Params:
        return get_params(self._cls)

    @cached.property
    def cls_params(self) -> Params | None:
        return self._cls.__dict__.get(PARAMS_ATTR)

    @cached.property
    def params_extras(self) -> ParamsExtras:
        return get_params_extras(self._cls)

    @cached.property
    def cls_params_extras(self) -> ParamsExtras | None:
        return (self.cls_metadata or {}).get(ParamsExtras)

    @cached.property
    def cls_metadata(self) -> Metadata | None:
        return self._cls.__dict__.get(METADATA_ATTR)

    @cached.property
    def merged_metadata(self) -> Metadata:
        return get_merged_metadata(self._cls)

    ##

    class _FoundFields(ta.NamedTuple):
        fields: dict[str, dc.Field]
        field_owners: dict[str, type]

    @lang.cached_function
    def _find_fields(self) -> _FoundFields:
        if self._constructing:
            check.in_(FIELDS_ATTR, self._cls.__dict__)

        fields: dict[str, dc.Field] = {}
        field_owners: dict[str, type] = {}

        for b in self._cls.__mro__[-1:0:-1]:
            base_fields = getattr(b, FIELDS_ATTR, None)
            if base_fields is not None:
                for name in reflect(b).cls_annotations:
                    try:
                        f = base_fields[name]
                    except KeyError:
                        continue
                    fields[f.name] = f
                    field_owners[f.name] = b

        cls_fields = getattr(self._cls, FIELDS_ATTR)
        for name, ann in self.cls_annotations.items():
            if is_kw_only(self._cls, ann):
                continue
            fields[name] = cls_fields[name]
            field_owners[name] = self._cls

        return ClassInfo._FoundFields(fields, field_owners)

    @cached.property
    def fields(self) -> ta.Mapping[str, dc.Field]:
        return self._find_fields().fields

    @cached.property
    def instance_fields(self) -> ta.Sequence[dc.Field]:
        return [f for f in self.fields.values() if field_type(f) is FieldType.INSTANCE]

    @cached.property
    def field_owners(self) -> ta.Mapping[str, type]:
        return self._find_fields().field_owners

    ##

    @cached.property
    def generic_mro(self) -> ta.Sequence[rfl.Type]:
        return rfl.ALIAS_UPDATING_GENERIC_SUBSTITUTION.generic_mro(self._cls)

    @cached.property
    def generic_mro_lookup(self) -> ta.Mapping[type, rfl.Type]:
        return col.make_map(((check.not_none(rfl.get_concrete_type(g)), g) for g in self.generic_mro), strict=True)

    @cached.property
    def generic_replaced_field_types(self) -> ta.Mapping[str, rfl.Type]:
        ret: dict[str, ta.Any] = {}
        for f in self.fields.values():
            fo = self.field_owners[f.name]
            go = self.generic_mro_lookup[fo]
            tvr = rfl.get_type_var_replacements(go)
            fty = rfl.type_(f.type)
            rty = rfl.replace_type_vars(fty, tvr, update_aliases=True)
            ret[f.name] = rty
        return ret

    @cached.property
    def generic_replaced_field_annotations(self) -> ta.Mapping[str, ta.Any]:
        return {k: rfl.to_annotation(v) for k, v in self.generic_replaced_field_types.items()}


##


_CLASS_INFO_CACHE: ta.MutableMapping[type, ClassInfo] = weakref.WeakKeyDictionary()


def reflect(obj: ta.Any) -> ClassInfo:
    cls = obj if isinstance(obj, type) else type(obj)
    try:
        return _CLASS_INFO_CACHE[cls]
    except KeyError:
        _CLASS_INFO_CACHE[cls] = info = ClassInfo(cls)
        return info



# Based on pytest-trio, licensed under the MIT license, duplicated below.
#
#  https://github.com/python-trio/pytest-trio/tree/cd6cc14b061d34f35980e38c44052108ed5402d1
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import functools
import sys
import typing as ta

import pytest
from _pytest.outcomes import Skipped  # noqa
from _pytest.outcomes import XFailed  # noqa

from ...... import lang
from .base import AsyncsBackend


if ta.TYPE_CHECKING:
    import trio
else:
    trio = lang.proxy_import('trio', extras=['abc'])


class TrioAsyncsBackend(AsyncsBackend):
    name = 'trio'

    def is_available(self) -> bool:
        return lang.can_import('trio')

    def is_imported(self) -> bool:
        return 'trio' in sys.modules

    #

    def wrap_runner(self, fn):
        @functools.wraps(fn)
        def wrapper(**kwargs):
            __tracebackhide__ = True

            clocks = {k: c for k, c in kwargs.items() if isinstance(c, trio.abc.Clock)}
            if not clocks:
                clock = None
            elif len(clocks) == 1:
                clock = list(clocks.values())[0]  # noqa
            else:
                raise ValueError(f'Expected at most one Clock in kwargs, got {clocks!r}')

            instruments = [i for i in kwargs.values() if isinstance(i, trio.abc.Instrument)]

            try:
                return trio.run(
                    functools.partial(fn, **kwargs),
                    clock=clock,
                    instruments=instruments,
                )

            except BaseExceptionGroup as eg:
                queue: list[BaseException] = [eg]
                leaves = []

                while queue:
                    ex = queue.pop()
                    if isinstance(ex, BaseExceptionGroup):
                        queue.extend(ex.exceptions)
                    else:
                        leaves.append(ex)

                if len(leaves) == 1:
                    if isinstance(leaves[0], XFailed):
                        pytest.xfail()
                    if isinstance(leaves[0], Skipped):
                        pytest.skip()

                # Since our leaf exceptions don't consist of exactly one 'magic' skipped or xfailed exception, re-raise
                # the whole group.
                raise

        return wrapper



import dataclasses as dc
import typing as ta

from ... import check
from ... import reflect as rfl
from ..base import MarshalContext
from ..base import Marshaler
from ..base import MarshalerFactory
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..base import UnmarshalerFactory
from ..values import Value


@dc.dataclass(frozen=True)
class LiteralMarshaler(Marshaler):
    e: Marshaler
    vs: frozenset

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        return self.e.marshal(ctx, check.in_(o, self.vs))


class LiteralMarshalerFactory(MarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Literal)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        lty = check.isinstance(rty, rfl.Literal)
        ety = check.single(set(map(type, lty.args)))
        return LiteralMarshaler(ctx.make(ety), frozenset(lty.args))


@dc.dataclass(frozen=True)
class LiteralUnmarshaler(Unmarshaler):
    e: Unmarshaler
    vs: frozenset

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        return check.in_(self.e.unmarshal(ctx, v), self.vs)


class LiteralUnmarshalerFactory(UnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Literal)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        lty = check.isinstance(rty, rfl.Literal)
        ety = check.single(set(map(type, lty.args)))
        return LiteralUnmarshaler(ctx.make(ety), frozenset(lty.args))



"""
TODO:
 - proxy_init 'as' alias support - attrs of (src, dst)
"""
import contextlib
import importlib.util
import sys
import types
import typing as ta


##


def can_import(name: str, package: str | None = None) -> bool:
    try:
        spec = importlib.util.find_spec(name, package)
    except ImportError:
        return False
    else:
        return spec is not None


##


def lazy_import(
        name: str,
        package: str | None = None,
        *,
        optional: bool = False,
        cache_failure: bool = False,
) -> ta.Callable[[], ta.Any]:
    result = not_set = object()

    def inner():
        nonlocal result

        if result is not not_set:
            if isinstance(result, Exception):
                raise result
            return result

        try:
            mod = importlib.import_module(name, package=package)

        except Exception as e:
            if optional:
                if cache_failure:
                    result = None
                return None

            if cache_failure:
                result = e
            raise

        result = mod
        return mod

    return inner


def proxy_import(
        name: str,
        package: str | None = None,
        extras: ta.Iterable[str] | None = None,
) -> types.ModuleType:
    if isinstance(extras, str):
        raise TypeError(extras)

    omod = None

    def __getattr__(att):  # noqa
        nonlocal omod
        if omod is None:
            omod = importlib.import_module(name, package=package)
            if extras:
                for x in extras:
                    importlib.import_module(f'{name}.{x}', package=package)
        return getattr(omod, att)

    lmod = types.ModuleType(name)
    lmod.__getattr__ = __getattr__  # type: ignore
    return lmod


##


SPECIAL_IMPORTABLE: ta.AbstractSet[str] = frozenset([
    '__init__.py',
    '__main__.py',
])


def yield_importable(
        package_root: str,
        *,
        recursive: bool = False,
        filter: ta.Callable[[str], bool] | None = None,  # noqa
        include_special: bool = False,
) -> ta.Iterator[str]:
    from importlib import resources

    def rec(cur):
        if cur.split('.')[-1] == '__pycache__':
            return

        try:
            module = sys.modules[cur]
        except KeyError:
            try:
                __import__(cur)
            except ImportError:
                return
            module = sys.modules[cur]

        # FIXME: pyox
        if getattr(module, '__file__', None) is None:
            return

        for file in resources.files(cur).iterdir():
            if file.is_file() and file.name.endswith('.py'):
                if not (include_special or file.name not in SPECIAL_IMPORTABLE):
                    continue

                name = cur + '.' + file.name[:-3]
                if filter is not None and not filter(name):
                    continue

                yield name

            elif recursive and file.is_dir():
                name = cur + '.' + file.name
                if filter is not None and not filter(name):
                    continue
                with contextlib.suppress(ImportError, NotImplementedError):
                    yield from rec(name)

    yield from rec(package_root)


def yield_import_all(
        package_root: str,
        *,
        globals: dict[str, ta.Any] | None = None,  # noqa
        locals: dict[str, ta.Any] | None = None,  # noqa
        recursive: bool = False,
        filter: ta.Callable[[str], bool] | None = None,  # noqa
        include_special: bool = False,
) -> ta.Iterator[str]:
    for import_path in yield_importable(
            package_root,
            recursive=recursive,
            filter=filter,
            include_special=include_special,
    ):
        __import__(import_path, globals=globals, locals=locals)
        yield import_path


def import_all(
        package_root: str,
        *,
        recursive: bool = False,
        filter: ta.Callable[[str], bool] | None = None,  # noqa
        include_special: bool = False,
) -> None:
    for _ in yield_import_all(
            package_root,
            recursive=recursive,
            filter=filter,
            include_special=include_special,
    ):
        pass


def try_import(spec: str) -> types.ModuleType | None:
    s = spec.lstrip('.')
    l = len(spec) - len(s)
    try:
        return __import__(s, globals(), level=l)
    except ImportError:
        return None


##


def resolve_import_name(name: str, package: str | None = None) -> str:
    level = 0

    if name.startswith('.'):
        if not package:
            raise TypeError("the 'package' argument is required to perform a relative import for {name!r}")
        for character in name:
            if character != '.':
                break
            level += 1

    name = name[level:]

    if not isinstance(name, str):
        raise TypeError(f'module name must be str, not {type(name)}')
    if level < 0:
        raise ValueError('level must be >= 0')
    if level > 0:
        if not isinstance(package, str):
            raise TypeError('__package__ not set to a string')
        elif not package:
            raise ImportError('attempted relative import with no known parent package')
    if not name and level == 0:
        raise ValueError('Empty module name')

    if level > 0:
        bits = package.rsplit('.', level - 1)  # type: ignore
        if len(bits) < level:
            raise ImportError('attempted relative import beyond top-level package')
        base = bits[0]
        name = f'{base}.{name}' if name else base

    return name


##


_REGISTERED_CONDITIONAL_IMPORTS: dict[str, list[str] | None] = {}


def _register_conditional_import(when: str, then: str, package: str | None = None) -> None:
    wn = resolve_import_name(when, package)
    tn = resolve_import_name(then, package)
    if tn in sys.modules:
        return
    if wn in sys.modules:
        __import__(tn)
    else:
        tns = _REGISTERED_CONDITIONAL_IMPORTS.setdefault(wn, [])
        if tns is None:
            raise Exception(f'Conditional import trigger already cleared: {wn=} {tn=}')
        tns.append(tn)


def _trigger_conditional_imports(package: str) -> None:
    tns = _REGISTERED_CONDITIONAL_IMPORTS.get(package, [])
    if tns is None:
        raise Exception(f'Conditional import trigger already cleared: {package=}')
    _REGISTERED_CONDITIONAL_IMPORTS[package] = None
    for tn in tns:
        __import__(tn)


##


class NamePackage(ta.NamedTuple):
    name: str
    package: str


class _ProxyInit:
    class _Import(ta.NamedTuple):
        pkg: str
        attr: str

    def __init__(
            self,
            name_package: NamePackage,
            *,
            globals: ta.MutableMapping[str, ta.Any] | None = None,  # noqa
            update_globals: bool = False,
    ) -> None:
        super().__init__()

        self._name_package = name_package
        self._globals = globals
        self._update_globals = update_globals

        self._imps_by_attr: dict[str, _ProxyInit._Import] = {}
        self._mods_by_pkgs: dict[str, ta.Any] = {}

    @property
    def name_package(self) -> NamePackage:
        return self._name_package

    def add(self, package: str, attrs: ta.Iterable[str | tuple[str, str]]) -> None:
        if isinstance(attrs, str):
            raise TypeError(attrs)
        for attr in attrs:
            if isinstance(attr, tuple):
                imp_attr, attr = attr
            else:
                imp_attr = attr
            self._imps_by_attr[attr] = self._Import(package, imp_attr)

    def get(self, attr: str) -> ta.Any:
        try:
            imp = self._imps_by_attr[attr]
        except KeyError:
            raise AttributeError(attr)  # noqa

        try:
            mod = self._mods_by_pkgs[imp.pkg]
        except KeyError:
            mod = importlib.import_module(imp.pkg, package=self._name_package.package)

        val = getattr(mod, imp.attr)

        if self._update_globals and self._globals is not None:
            self._globals[attr] = val

        return val


def proxy_init(
        globals: ta.MutableMapping[str, ta.Any],  # noqa
        package: str,
        attrs: ta.Iterable[str | tuple[str, str]],
) -> None:
    if isinstance(attrs, str):
        raise TypeError(attrs)

    init_name_package = NamePackage(
        globals['__name__'],
        globals['__package__'],
    )

    pi: _ProxyInit
    try:
        pi = globals['__proxy_init__']
    except KeyError:
        pi = _ProxyInit(
            init_name_package,
            globals=globals,
        )
        globals['__proxy_init__'] = pi
        globals['__getattr__'] = pi.get
    else:
        if pi.name_package != init_name_package:
            raise Exception(f'Wrong init name: {pi.name_package=} != {init_name_package=}')

    pi.add(package, attrs)


##


def get_real_module_name(globals: ta.Mapping[str, ta.Any]) -> str:  # noqa
    module = sys.modules[globals['__name__']]

    if module.__spec__ and module.__spec__.name:
        return module.__spec__.name

    if module.__package__:
        return module.__package__

    raise RuntimeError("Can't determine real module name")



# ruff: noqa: UP006 UP007
# @omlish-lite
import abc
import concurrent.futures as cf
import dataclasses as dc
import logging
import socket
import typing as ta

from ..addresses import SocketAndAddress
from ..handlers import SocketHandler
from ..io import SocketIoPair
from ..io import close_socket_immediately


SocketServerHandler = ta.Callable[['SocketAndAddress'], None]  # ta.TypeAlias


##


class SocketServerHandler_(abc.ABC):  # noqa
    @abc.abstractmethod
    def __call__(self, conn: SocketAndAddress) -> None:
        raise NotImplementedError


##


@dc.dataclass(frozen=True)
class StandardSocketServerHandler(SocketServerHandler_):
    handler: SocketServerHandler

    timeout: ta.Optional[float] = None

    # http://bugs.python.org/issue6192
    # TODO: https://eklitzke.org/the-caveats-of-tcp-nodelay
    disable_nagle_algorithm: bool = False

    no_close: bool = False

    def __call__(self, conn: SocketAndAddress) -> None:
        try:
            if self.timeout is not None:
                conn.socket.settimeout(self.timeout)

            if self.disable_nagle_algorithm:
                conn.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)

            self.handler(conn)

        finally:
            close_socket_immediately(conn.socket)


#


@dc.dataclass(frozen=True)
class CallbackWrappedSocketServerHandler(SocketServerHandler_):
    handler: SocketServerHandler

    before_handle: ta.Optional[SocketServerHandler] = None
    after_handle: ta.Optional[SocketServerHandler] = None

    # Return True if suppress like __exit__
    on_error: ta.Optional[ta.Callable[[SocketAndAddress, Exception], bool]] = None

    finally_: ta.Optional[SocketServerHandler] = None

    def __call__(self, conn: SocketAndAddress) -> None:
        try:
            if (before_handle := self.before_handle) is not None:
                before_handle(conn)

            self.handler(conn)

        except Exception as e:
            if (on_error := self.on_error) is not None and on_error(conn, e):
                pass
            else:
                raise

        else:
            if (after_handle := self.after_handle) is not None:
                after_handle(conn)

        finally:
            if (finally_ := self.finally_) is not None:
                finally_(conn)


#


@dc.dataclass(frozen=True)
class SocketHandlerSocketServerHandler(SocketServerHandler_):
    handler: SocketHandler

    r_buf_size: int = -1
    w_buf_size: int = 0

    def __call__(self, conn: SocketAndAddress) -> None:
        fp = SocketIoPair.from_socket(
            conn.socket,
            r_buf_size=self.r_buf_size,
            w_buf_size=self.w_buf_size,
        )

        self.handler(conn.address, fp)


#


@dc.dataclass(frozen=True)
class SocketWrappingSocketServerHandler(SocketServerHandler_):
    handler: SocketServerHandler
    wrapper: ta.Callable[[SocketAndAddress], SocketAndAddress]

    def __call__(self, conn: SocketAndAddress) -> None:
        wrapped_conn = self.wrapper(conn)
        self.handler(wrapped_conn)


#

@dc.dataclass(frozen=True)
class ExecutorSocketServerHandler(SocketServerHandler_):
    handler: SocketServerHandler
    executor: cf.Executor

    def __call__(self, conn: SocketAndAddress) -> None:
        self.executor.submit(self.handler, conn)


#


@dc.dataclass(frozen=True)
class ExceptionLoggingSocketServerHandler(SocketServerHandler_):
    handler: SocketServerHandler
    log: logging.Logger

    ignored: ta.Optional[ta.Container[ta.Type[Exception]]] = None

    def __call__(self, conn: SocketAndAddress) -> None:
        try:
            return self.handler(conn)

        except Exception as e:  # noqa
            if (ignored := self.ignored) is None or type(e) not in ignored:
                self.log.exception('Error in handler %r for conn %r', self.handler, conn)

            raise



# ruff: noqa: UP006 UP007
# @omlish-lite
import asyncio
import typing as ta

from ...lite.timeouts import Timeout
from ...lite.timeouts import TimeoutLike


async def asyncio_wait_until_can_connect(
        host: ta.Any = None,
        port: ta.Any = None,
        *,
        timeout: ta.Optional[TimeoutLike] = None,
        on_fail: ta.Optional[ta.Callable[[BaseException], None]] = None,
        sleep_s: float = .1,
        exception: ta.Union[ta.Type[BaseException], ta.Tuple[ta.Type[BaseException], ...]] = (Exception,),
) -> None:
    timeout = Timeout.of(timeout)

    async def inner():
        while True:
            timeout()

            try:
                reader, writer = await asyncio.open_connection(host, port)

            except asyncio.CancelledError:
                raise

            except exception as e:  # noqa
                if on_fail is not None:
                    on_fail(e)

            else:
                writer.close()
                await asyncio.wait_for(writer.wait_closed(), timeout=timeout.or_(None))
                break

            await asyncio.sleep(min(sleep_s, timeout.remaining()))

    if timeout() != float('inf'):
        await asyncio.wait_for(inner(), timeout=timeout())
    else:
        await inner()



"""
Dynamically scoped variables (implemented by stackwalking). Unlike threadlocals these are generator-correct both in
binding and retrieval, and unlike ContextVars they require no manual context management. They are however *slow* and
should be used sparingly (once per sql statement executed not once per inner function call).

TODO:
 - clj-style binding conveyance
 - contextvar/async interop
 - 'partializer'
"""
import contextlib
import functools
import sys
import types
import typing as ta
import weakref

from . import lang


T = ta.TypeVar('T')


_HOISTED_CODE_DEPTH: ta.MutableMapping[types.CodeType, int] = weakref.WeakKeyDictionary()
_MAX_HOIST_DEPTH = 0


def hoist(depth=0):  # noqa
    def inner(fn):
        _HOISTED_CODE_DEPTH[fn.__code__] = depth
        global _MAX_HOIST_DEPTH
        _MAX_HOIST_DEPTH = max(_MAX_HOIST_DEPTH, depth)
        return fn
    return inner


hoist()(contextlib.ExitStack.enter_context)  # noqa


class MISSING(lang.Marker):
    pass


class UnboundVarError(ValueError):
    pass


class Var(ta.Generic[T]):
    def __init__(
            self,
            default: type[MISSING] | T = MISSING,
            *,
            new: ta.Callable[[], T] | type[MISSING] = MISSING,
            validate: ta.Callable[[T], None] | None = None,
    ) -> None:
        super().__init__()

        if default is not MISSING and new is not MISSING:
            raise TypeError('Cannot set both default and new')
        elif default is not MISSING:
            new = lambda: default  # type: ignore
        self._new: type[MISSING] | ta.Callable[[], T] = new
        self._validate = validate
        self._bindings_by_frame: ta.MutableMapping[types.FrameType, ta.MutableMapping[int, Binding]] = weakref.WeakValueDictionary()  # noqa

    @ta.overload
    def __call__(self) -> T:
        ...

    @ta.overload
    def __call__(self, value: T, **kwargs: ta.Any) -> ta.ContextManager[T]:
        ...

    def __call__(self, *args, **kwargs):
        if not args:
            if kwargs:
                raise TypeError(kwargs)
            return self.value
        elif len(args) == 1:
            return self.binding(*args, **kwargs)
        else:
            raise TypeError(args)

    def binding(self, value: T, *, offset: int = 0) -> ta.ContextManager[T]:
        if self._validate is not None:
            self._validate(self.value)
        return Binding(self, value, offset=offset)

    def with_binding(self, value):  # noqa
        def outer(fn):
            @functools.wraps(fn)
            def inner(*args, **kwargs):
                with self.binding(value):
                    return fn(*args, **kwargs)
            return inner
        return outer

    def with_binding_fn(self, binding_fn):  # noqa
        this = self

        def outer(fn):
            class Descriptor:

                @staticmethod
                @functools.wraps(fn)
                def __call__(*args, **kwargs):  # noqa
                    with this.binding(binding_fn(*args, **kwargs)):
                        return fn(*args, **kwargs)

                def __get__(self, obj, cls=None):
                    bound_binding_fn = binding_fn.__get__(obj, cls)
                    bound_fn = fn.__get__(obj, cls)

                    @functools.wraps(fn)
                    def inner(*args, **kwargs):
                        with this.binding(bound_binding_fn(*args, **kwargs)):
                            return bound_fn(*args, **kwargs)

                    return inner

            dct: dict[str, ta.Any] = {k: getattr(fn, k) for k in functools.WRAPPER_ASSIGNMENTS}
            return lang.new_type(fn.__name__, (Descriptor,), dct)()

        return outer

    @property
    def values(self) -> ta.Iterator[T]:
        frame = sys._getframe().f_back  # noqa
        while frame:
            try:
                frame_bindings = self._bindings_by_frame[frame]
            except KeyError:
                pass
            else:
                for level, frame_binding in sorted(frame_bindings.items()):  # noqa
                    yield frame_binding._value  # noqa
            frame = frame.f_back

        if self._new is not MISSING:
            yield self._new()  # type: ignore

    def __iter__(self) -> ta.Iterator[T]:
        return self.values

    @property
    def value(self) -> T:
        try:
            return next(self.values)
        except StopIteration:
            raise UnboundVarError from None


class Binding(ta.Generic[T]):
    _frame: types.FrameType
    _frame_bindings: ta.MutableMapping[int, 'Binding']
    _level: int

    def __init__(self, var: Var[T], value: T, *, offset: int = 0) -> None:
        super().__init__()

        self._var = var
        self._value = value
        self._offset = offset

    def __enter__(self) -> T:
        frame = sys._getframe(self._offset).f_back  # noqa
        lag_frame: types.FrameType | None = frame
        while lag_frame is not None:
            for cur_depth in range(_MAX_HOIST_DEPTH + 1):
                if lag_frame is None:
                    break
                try:
                    lag_hoist = _HOISTED_CODE_DEPTH[lag_frame.f_code]
                except KeyError:
                    pass
                else:
                    if lag_hoist >= cur_depth:
                        frame = lag_frame = lag_frame.f_back
                        break
                lag_frame = lag_frame.f_back
            else:
                break

        if frame is None:
            raise RuntimeError
        self._frame = frame
        try:
            self._frame_bindings = self._var._bindings_by_frame[self._frame]  # noqa
        except KeyError:
            self._frame_bindings = self._var._bindings_by_frame[self._frame] = weakref.WeakValueDictionary()  # noqa
            self._level = 0
        else:
            self._level = min(self._frame_bindings.keys() or [1]) - 1

        self._frame_bindings[self._level] = self
        return self._value

    def __exit__(self, et, e, tb):
        if self._frame_bindings[self._level] is not self:
            raise TypeError

        del self._frame_bindings[self._level]
        del self._frame_bindings
        del self._frame


class _GeneratorContextManager(contextlib._GeneratorContextManager):  # noqa
    @hoist(2)
    def __enter__(self):
        return super().__enter__()


def contextmanager(fn):  # noqa
    @functools.wraps(fn)
    def helper(*args, **kwds):
        return _GeneratorContextManager(fn, args, kwds)
    return helper  # noqa



import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang


ConfigurableConfigT = ta.TypeVar('ConfigurableConfigT', bound='Configurable.Config')


class Configurable(ta.Generic[ConfigurableConfigT], lang.Abstract):
    @dc.dataclass(frozen=True, kw_only=True)
    class Config:
        """Does not use any dc metaclasses to preserve typechecking."""

        configurable_cls: ta.ClassVar[type['Configurable']]

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if not lang.is_abstract_class(cls):
            check.in_('Config', cls.__dict__)
            cfg_cls = check.issubclass(cls.Config, Configurable.Config)
            check.not_in('configurable_cls', cfg_cls.__dict__)
            check.state(dc.is_immediate_dataclass(cfg_cls))
            cfg_cls.configurable_cls = cls

    def __init__(self, config: ConfigurableConfigT) -> None:
        super().__init__()

        self._config: ConfigurableConfigT = check.isinstance(config, self.Config)  # type: ignore[assignment]

    @property
    def config(self) -> ConfigurableConfigT:
        return self._config



import decimal
import fractions
import typing as ta

from ... import check
from ..base import MarshalContext
from ..base import Marshaler
from ..base import TypeMapMarshalerFactory
from ..base import TypeMapUnmarshalerFactory
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..values import Value


class ComplexMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: complex) -> Value:
        return [o.real, o.imag]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        real, imag = check.isinstance(v, list)
        return complex(real, imag)


class DecimalMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: decimal.Decimal) -> Value:
        return str(o)

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return decimal.Decimal(check.isinstance(v, str))


class FractionMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: fractions.Fraction) -> Value:
        return [o.numerator, o.denominator]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        num, denom = check.isinstance(v, list)
        return fractions.Fraction(num, denom)


NUMBERS_MARSHALER_FACTORY = TypeMapMarshalerFactory({
    complex: ComplexMarshalerUnmarshaler(),
    decimal.Decimal: DecimalMarshalerUnmarshaler(),
    fractions.Fraction: FractionMarshalerUnmarshaler(),
})

NUMBERS_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({
    complex: ComplexMarshalerUnmarshaler(),
    decimal.Decimal: DecimalMarshalerUnmarshaler(),
    fractions.Fraction: FractionMarshalerUnmarshaler(),
})



import typing as ta

from ... import dataclasses as dc
from ... import lang
from .base import Node
from .exprs import CanExpr
from .exprs import Expr
from .exprs import ExprBuilder
from .idents import Ident
from .relations import CanRelation
from .relations import Relation
from .relations import RelationBuilder
from .stmts import Stmt


##


class SelectItem(Node, lang.Final):
    v: Expr
    a: Ident | None = dc.xfield(None, repr_fn=dc.opt_repr)


class Select(Stmt, lang.Final):
    items: ta.Sequence[SelectItem] = dc.xfield(coerce=tuple)
    from_: Relation | None = dc.xfield(None, repr_fn=dc.opt_repr)
    where: Expr | None = dc.xfield(None, repr_fn=dc.opt_repr)


CanSelectItem: ta.TypeAlias = SelectItem | CanExpr


class SelectBuilder(RelationBuilder, ExprBuilder):
    def select_item(self, o: CanSelectItem) -> SelectItem:
        if isinstance(o, SelectItem):
            return o
        else:
            return SelectItem(self.expr(o))

    def select(
            self,
            items: ta.Sequence[CanSelectItem],
            from_: CanRelation | None = None,
            where: CanExpr | None = None,
    ) -> Select:
        return Select(
            tuple(self.select_item(i) for i in items),
            from_=self.relation(from_) if from_ is not None else None,
            where=self.expr(where) if where is not None else None,
        )



import dataclasses as dc
import enum
import typing as ta

from ... import check
from ... import lang


##


class QueryMode(enum.Enum):
    QUERY = enum.auto()
    EXEC = enum.auto()


@dc.dataclass(frozen=True)
class Query(lang.Final):
    mode: QueryMode
    text: str
    args: ta.Sequence[ta.Any]

    #

    @classmethod
    @ta.overload
    def of(
            cls,
            query: 'Query',
    ) -> 'Query':
        ...

    @classmethod
    @ta.overload
    def of(
            cls,
            text: str,
            *args: ta.Any,
            mode: str | QueryMode = QueryMode.QUERY,
    ) -> 'Query':
        ...

    @classmethod  # type: ignore[misc]
    def of(cls, obj, *args, **kwargs):
        if isinstance(obj, Query):
            check.arg(not args)
            check.arg(not kwargs)
            return obj

        elif isinstance(obj, str):
            mode = kwargs.pop('mode', QueryMode.QUERY)
            if isinstance(mode, str):
                mode = QueryMode[mode.upper()]
            check.arg(not kwargs)

            return cls(
                mode=mode,
                text=obj,
                args=args,
            )

        else:
            raise TypeError(obj)



import typing as ta

from .base import Querier
from .base import Rows
from .queries import Query
from .queries import QueryMode


##


@ta.overload
def query(
        querier: Querier,
        query: Query,  # noqa
) -> Rows:
    ...


@ta.overload
def query(
        querier: Querier,
        text: str,
        *args: ta.Any,
) -> Rows:
    ...


def query(
        querier,
        obj,
        *args,
):
    if isinstance(obj, Query):
        q = obj
    else:
        q = Query.of(obj, *args, mode=QueryMode.QUERY)

    return querier.query(q)


##


@ta.overload
def exec(  # noqa
        querier: Querier,
        query: Query,  # noqa
) -> None:
    ...


@ta.overload
def exec(  # noqa
        querier: Querier,
        text: str,
        *args: ta.Any,
) -> None:
    ...


def exec(  # noqa
    querier,
    obj,
    *args,
):
    if isinstance(obj, Query):
        q = obj
    else:
        q = Query.of(obj, *args, mode=QueryMode.EXEC)

    with querier.query(q):
        pass



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
usecase: supervisor process groups
"""
import typing as ta

from ...lite.check import check
from ..types import ConfigMap


##


def build_config_named_children(
        o: ta.Union[
            ta.Sequence[ConfigMap],
            ta.Mapping[str, ConfigMap],
            None,
        ],
        *,
        name_key: str = 'name',
) -> ta.Optional[ta.Sequence[ConfigMap]]:
    if o is None:
        return None

    lst: ta.List[ConfigMap] = []
    if isinstance(o, ta.Mapping):
        for k, v in o.items():
            check.isinstance(v, ta.Mapping)
            if name_key in v:
                n = v[name_key]
                if k != n:
                    raise KeyError(f'Given names do not match: {n} != {k}')
                lst.append(v)
            else:
                lst.append({name_key: k, **v})

    else:
        check.not_isinstance(o, str)
        lst.extend(o)

    seen = set()
    for d in lst:
        n = d['name']
        if n in d:
            raise KeyError(f'Duplicate name: {n}')
        seen.add(n)

    return lst



from dataclasses import (  # noqa
    FrozenInstanceError,

    MISSING,
    KW_ONLY,

    InitVar,
    Field,

    field,

    dataclass,
    make_dataclass,

    fields,

    is_dataclass,

    # asdict,
    # astuple,

    # replace,
)

from .impl.api import (  # noqa
    field as xfield,

    dataclass as xdataclass,
    make_dataclass as xmake_dataclass,

    extra_params,
)

from .impl.as_ import (  # noqa
    asdict,
    astuple,
)

from .impl.params import (  # noqa
    FieldExtras,
    get_field_extras,
)

from .impl.replace import (  # noqa
    replace,
)


##


globals()['field'] = xfield

globals()['dataclass'] = xdataclass
globals()['make_dataclass'] = xmake_dataclass


##


from .impl.exceptions import (  # noqa
    FieldValidationError,
    ValidationError,
)

from .impl.metaclass import (  # noqa
    DataMeta,
    Data,
    Frozen,
    Case,
    Box,
)

from .impl.metadata import (  # noqa
    Metadata,

    get_merged_metadata,

    UserMetadata,
    metadata,

    Validate,
    validate,

    Init,
    init,
)

from .impl.reflect import (  # noqa
    ClassInfo,
    reflect,
)

from .static import (  # noqa
    Static,
)

from .utils import (  # noqa
    opt_repr,
    truthy_repr,

    fields_dict,
    field_modifier,
    chain_metadata,
    update_class_metadata,
    update_field_metadata,
    update_field_extras,
    update_fields,
    update_fields_metadata,

    shallow_astuple,
    shallow_asdict,

    deep_replace,

    iter_items,
    iter_keys,
    iter_values,
)

##

from ..lite.dataclasses import (  # noqa
    is_immediate_dataclass,

    dataclass_maybe_post_init as maybe_post_init,
)



import os
import typing as ta
import weakref

from ... import check
from ..forkhooks import ForkHook
from ..forkhooks import ProcessOriginTracker
from .base import BaseDeathpact


##


class PipeDeathpact(BaseDeathpact):
    """
    NOTE: Closes write side in children lazily on poll - does not proactively close write sides on fork. This means
          parents which fork children into codepaths unaware of live PipeDeathpacts will leave write sides open in those
          children, potentially leading to zombies (if those children outlast the parent). Use ForkAwarePipeDeathpact to
          handle such cases.
    """

    def __init__(self, **kwargs: ta.Any) -> None:
        super().__init__(**kwargs)

        self._rfd: int | None = None
        self._wfd: int | None = None

        self._process_origin = ProcessOriginTracker()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(rfd={self._rfd}, wfd={self._wfd})'

    @property
    def pass_fd(self) -> int:
        return check.not_none(self._rfd)

    def is_parent(self) -> bool:
        return self._process_origin.is_in_origin_process()

    #

    def __enter__(self) -> ta.Self:
        check.state(self.is_parent())

        check.none(self._rfd)
        check.none(self._wfd)

        self._rfd, self._wfd = os.pipe()

        os.set_blocking(self._rfd, False)

        return self

    def close(self) -> None:
        if self._rfd is not None:
            os.close(self._rfd)
            self._rfd = None

        if self._wfd is not None:
            if self.is_parent():
                os.close(check.not_none(self._wfd))
            self._wfd = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _close_wfd_if_not_parent(self) -> None:
        if self._wfd is not None:
            if not self.is_parent():
                os.close(check.not_none(self._wfd))
            self._wfd = None

    #

    def __getstate__(self):
        return {
            **self.__dict__,
            **dict(
                _wfd=None,
            ),
        }

    def __setstate__(self, state):
        self.__dict__.update(state)

    #

    def should_die(self) -> bool:
        self._close_wfd_if_not_parent()

        try:
            buf = os.read(check.not_none(self._rfd), 1)
        except BlockingIOError:
            return False

        if buf:
            self._print(f'Read data from pipe! This should not happen! Process state corrupt!')
            self.die()

        return True


#


class ForkAwarePipeDeathpact(PipeDeathpact):
    """
    TODO:
     - Despite no correct way to do threads+forks, still audit thread-safety. Is WeakSet threadsafe? Probably not..
    """

    _PARENTS: ta.ClassVar[ta.MutableSet['ForkAwarePipeDeathpact']] = weakref.WeakSet()

    def __init__(self, **kwargs: ta.Any) -> None:
        super().__init__(**kwargs)

        self._ForkHook.install()
        self._PARENTS.add(self)

    class _ForkHook(ForkHook):
        @classmethod
        def _after_fork_in_child(cls) -> None:
            for pdp in ForkAwarePipeDeathpact._PARENTS:
                pdp._close_wfd_if_not_parent()  # noqa
            ForkAwarePipeDeathpact._PARENTS.clear()



import typing as ta

from ... import dataclasses as dc
from ... import lang


##


Value: ta.TypeAlias = ta.Any


##


class Node(dc.Frozen, lang.Abstract, cache_hash=True):
    pass


class Builder(lang.Abstract):
    pass



"""
TODO:
 - https://github.com/ergochat/irctest/blob/master/irctest/irc_utils/message_parser.py ?
"""
import dataclasses as dc
import typing as ta


@dc.dataclass(frozen=True)
class Message:
    source: str | None
    command: str
    params: ta.Sequence[str]

    force_trailing: bool = False

    tags: ta.Mapping[str, str] | None = None
    client_only_tags: ta.Mapping[str, str] | None = None



import dataclasses as dc
import typing as ta

from ... import lang
from ..naming import Naming
from ..naming import translate_name
from ..registries import RegistryItem


##


class TypeTagging(RegistryItem, lang.Abstract, lang.Sealed):
    pass


class WrapperTypeTagging(TypeTagging, lang.Final):
    pass


@dc.dataclass(frozen=True)
class FieldTypeTagging(TypeTagging, lang.Final):
    field: str


##


@dc.dataclass(frozen=True)
class Impl:
    ty: type
    tag: str
    alts: ta.AbstractSet[str] = frozenset()


class Impls(ta.Sequence[Impl]):
    def __init__(
            self,
            lst: ta.Iterable[Impl],
    ) -> None:
        super().__init__()
        self._lst = list(lst)

        by_ty: dict[type, Impl] = {}
        by_tag: dict[str, Impl] = {}
        for i in self._lst:
            if i.ty in by_ty:
                raise TypeError(i.ty)
            if i.tag in by_tag:
                raise NameError(i.tag)
            for a in i.alts:
                if a in by_tag:
                    raise NameError(a)
            by_ty[i.ty] = i
            by_tag[i.tag] = i
            for a in i.alts:
                by_tag[a] = i
        self._by_ty = by_ty
        self._by_tag = by_tag

    def __iter__(self) -> ta.Iterator[Impl]:
        return iter(self._lst)

    def __len__(self) -> int:
        return len(self._lst)

    @ta.overload
    def __getitem__(self, index: int) -> Impl: ...

    @ta.overload
    def __getitem__(self, index: slice) -> ta.Sequence[Impl]: ...

    def __getitem__(self, index):
        return self._lst[index]

    @property
    def by_ty(self) -> ta.Mapping[type, Impl]:
        return self._by_ty

    @property
    def by_tag(self) -> ta.Mapping[str, Impl]:
        return self._by_tag


class Polymorphism:
    def __init__(
            self,
            ty: type,
            impls: ta.Iterable[Impl],
    ) -> None:
        super().__init__()
        self._ty = ty
        self._impls = Impls(impls)

        for i in self._impls:
            if not issubclass(i.ty, ty):
                raise TypeError(i.ty, ty)

    @property
    def ty(self) -> type:
        return self._ty

    @property
    def impls(self) -> Impls:
        return self._impls


def polymorphism_from_subclasses(
        ty: type,
        *,
        naming: Naming | None = None,
        strip_suffix: bool = False,
) -> Polymorphism:
    dct: dict[str, Impl] = {}

    seen: set[type] = set()
    todo: list[type] = [ty]
    while todo:
        cur = todo.pop()
        seen.add(cur)

        todo.extend(nxt for nxt in cur.__subclasses__() if nxt not in seen)

        if lang.is_abstract_class(cur):
            continue

        name = cur.__name__
        if strip_suffix:
            name = lang.strip_suffix(name, ty.__name__)
        if naming is not None:
            name = translate_name(name, naming)
        if name in dct:
            raise KeyError(f'Duplicate name: {name}')

        dct[name] = Impl(
            cur,
            name,
        )

    return Polymorphism(ty, dct.values())



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
TODO:
 - codification of https://docs.python.org/3/library/socket.html#socket-families
"""
import dataclasses as dc
import socket as socket_
import typing as ta


SocketAddress = ta.Any


##


@dc.dataclass(frozen=True)
class SocketAddressInfoArgs:
    host: ta.Optional[str]
    port: ta.Union[str, int, None]
    family: socket_.AddressFamily = socket_.AddressFamily.AF_UNSPEC
    type: int = 0
    proto: int = 0
    flags: socket_.AddressInfo = socket_.AddressInfo(0)


@dc.dataclass(frozen=True)
class SocketAddressInfo:
    family: socket_.AddressFamily
    type: int
    proto: int
    canonname: ta.Optional[str]
    sockaddr: SocketAddress


class SocketFamilyAndAddress(ta.NamedTuple):
    family: socket_.AddressFamily
    address: SocketAddress


def get_best_socket_family(
        host: ta.Optional[str],
        port: ta.Union[str, int, None],
        family: ta.Union[int, socket_.AddressFamily] = socket_.AddressFamily.AF_UNSPEC,
) -> SocketFamilyAndAddress:
    """https://github.com/python/cpython/commit/f289084c83190cc72db4a70c58f007ec62e75247"""

    infos = socket_.getaddrinfo(
        host,
        port,
        family,
        type=socket_.SOCK_STREAM,
        flags=socket_.AI_PASSIVE,
    )
    ai = SocketAddressInfo(*next(iter(infos)))
    return SocketFamilyAndAddress(ai.family, ai.sockaddr)


class SocketAndAddress(ta.NamedTuple):
    socket: socket_.socket
    address: SocketAddress

    def wrap_socket(self, fn: ta.Callable[[socket_.socket], socket_.socket]):
        return self._replace(socket=fn(self.socket))

    @classmethod
    def socket_wrapper(
            cls,
            fn: ta.Callable[[socket_.socket], socket_.socket],
    ) -> ta.Callable[['SocketAndAddress'], 'SocketAndAddress']:
        def inner(conn):
            return conn.wrap_socket(fn)
        return inner



# ruff: noqa: I001
# flake8: noqa: F401

from ._runtime.BufferedTokenStream import (  # type: ignore
    TokenStream,
)

from ._runtime.CommonTokenStream import (  # type: ignore
    CommonTokenStream,
)

from ._runtime.FileStream import (  # type: ignore
    FileStream,
)

from ._runtime.InputStream import (  # type: ignore
    InputStream,
)

from ._runtime.Lexer import (  # type: ignore
    Lexer,
)

from ._runtime.Parser import (  # type: ignore
    Parser,
)

from ._runtime.ParserRuleContext import (  # type: ignore
    ParserRuleContext,
    RuleContext,
)

from ._runtime.PredictionContext import (  # type: ignore
    PredictionContextCache,
)

from ._runtime.StdinStream import (  # type: ignore
    StdinStream,
)

from ._runtime.Token import (  # type: ignore
    Token,
)

from ._runtime.Utils import (  # type: ignore
    str_list,
)

from ._runtime.atn.ATN import (  # type: ignore
    ATN,
)

from ._runtime.atn.ATNDeserializer import (  # type: ignore
    ATNDeserializer,
)

from ._runtime.atn.LexerATNSimulator import (  # type: ignore
    LexerATNSimulator,
)

from ._runtime.atn.ParserATNSimulator import (  # type: ignore
    ParserATNSimulator,
)

from ._runtime.atn.PredictionMode import (  # type: ignore
    PredictionMode,
)

from ._runtime.dfa.DFA import (  # type: ignore
    DFA,
)

from ._runtime.error.DiagnosticErrorListener import (  # type: ignore
    DiagnosticErrorListener,
)

from ._runtime.error.ErrorListener import (  # type: ignore
    ErrorListener,
)

from ._runtime.error.ErrorStrategy import (  # type: ignore
    BailErrorStrategy,
)

from ._runtime.error.Errors import (  # type: ignore
    LexerNoViableAltException,
)

from ._runtime.error.Errors import (  # type: ignore
    IllegalStateException,
    NoViableAltException,
    RecognitionException,
)

from ._runtime.tree.Tree import (  # type: ignore
    ErrorNode,
    ParseTreeListener,
    ParseTreeVisitor,
    ParseTreeWalker,
    RuleNode,
    TerminalNode,
)



import dataclasses as dc
import typing as ta

from ... import check
from ... import reflect as rfl
from ..base import MarshalContext
from ..base import Marshaler
from ..base import MarshalerFactory
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..base import UnmarshalerFactory
from ..values import Value


@dc.dataclass(frozen=True)
class OptionalMarshaler(Marshaler):
    e: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        if o is None:
            return None
        return self.e.marshal(ctx, o)


class OptionalMarshalerFactory(MarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Union) and rty.is_optional

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return OptionalMarshaler(ctx.make(check.isinstance(rty, rfl.Union).without_none()))


@dc.dataclass(frozen=True)
class OptionalUnmarshaler(Unmarshaler):
    e: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        if v is None:
            return None
        return self.e.unmarshal(ctx, v)


class OptionalUnmarshalerFactory(UnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Union) and rty.is_optional

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return OptionalUnmarshaler(ctx.make(check.isinstance(rty, rfl.Union).without_none()))



import typing as ta

from ... import lang
from .base import Node
from .exprs import CanExpr
from .exprs import ExprBuilder


##


class Stmt(Node, lang.Abstract):
    pass


class ExprStmt(Stmt, lang.Final):
    pass


CanStmt: ta.TypeAlias = Stmt | CanExpr


class StmtBuilder(ExprBuilder):
    def stmt(self, o: CanStmt) -> Stmt:
        if isinstance(o, Stmt):
            return o
        else:
            return ExprStmt(self.expr(o))



import abc
import functools
import socket
import typing as ta

from .. import dataclasses as dc


##


class Wait(dc.Case):
    pass


class Waiter(abc.ABC):
    @abc.abstractmethod
    def do_wait(self) -> bool:
        raise NotImplementedError


@functools.singledispatch
def waiter_for(wait: Wait) -> Waiter:
    raise TypeError(wait)


##


class SequentialWait(Wait):
    waits: ta.Sequence[Wait]


class SequentialWaiter(Waiter):
    def __init__(self, waiters: ta.Sequence[Waiter]) -> None:
        super().__init__()

        self._waiters = waiters
        self._idx = 0

    def do_wait(self) -> bool:
        while self._idx < len(self._waiters):
            if not self._waiters[self._idx].do_wait():
                return False
            self._idx += 1
        return True


@waiter_for.register
def _(wait: SequentialWait) -> SequentialWaiter:
    return SequentialWaiter([waiter_for(c) for c in wait.waits])


##


class FnWait(Wait):
    fn: ta.Callable[[], bool]


class FnWaiter(Waiter, dc.Frozen):
    wait: FnWait

    def do_wait(self) -> bool:
        return self.wait.fn()


@waiter_for.register
def _(wait: FnWait) -> FnWaiter:
    return FnWaiter(wait)


##


class ConnectWait(Wait):
    address: ta.Any


class ConnectWaiter(Waiter, dc.Frozen):
    wait: ConnectWait

    def do_wait(self) -> bool:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.connect(self.wait.address)
            except ConnectionRefusedError:
                return False
            else:
                return True


@waiter_for.register
def _(wait: ConnectWait) -> ConnectWaiter:
    return ConnectWaiter(wait)



import functools
import typing as ta


##


def get_bit(bit: int, value: int) -> int:
    return (value >> bit) & 1


def get_bits(bits_from: int, num_bits: int, value: int) -> int:
    return (value & ((1 << (bits_from + num_bits)) - 1)) >> bits_from


def set_bit(bit: int, bit_value: int, value: int) -> int:
    if bit_value:
        return value | (1 << bit)
    else:
        return value & ~(1 << bit)


def set_bits(bits_from: int, num_bits: int, bits_value: int, value: int) -> int:
    return value & ~(((1 << num_bits) - 1) << bits_from) | (bits_value << bits_from)


##


def _gen_scalar_proxy_method(name):
    def inner(self, *args, **kwargs):
        return self.__class__(orig(self, *args, **kwargs))

    orig = getattr(int, name)
    return functools.wraps(orig)(inner)


def _gen_tuple_proxy_method(name):
    def inner(self, *args, **kwargs):
        return tuple(map(self.__class__, orig(self, *args, **kwargs)))

    orig = getattr(int, name)
    return functools.wraps(orig)(inner)


class FixedWidthInt(int):
    BITS: ta.ClassVar[int]
    SIGNED: ta.ClassVar[bool]

    MIN: ta.ClassVar[int]
    MAX: ta.ClassVar[int]

    MASK: ta.ClassVar[int]

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if not isinstance(cls.BITS, int):
            raise TypeError(cls.BITS)

        if cls.SIGNED:
            cls.MIN = -(1 << (cls.BITS - 1))
            cls.MAX = (1 << (cls.BITS - 1)) - 1
        else:
            cls.MIN = 0
            cls.MAX = (1 << cls.BITS) - 1

        cls.MASK = (1 << cls.BITS) - 1

    @classmethod
    def clamp(cls, value: int) -> int:
        return ((value - cls.MIN) & cls.MASK) + cls.MIN

    def __new__(cls, value: int) -> 'FixedWidthInt':
        return super().__new__(cls, cls.clamp(value))  # noqa

    SCALAR_PROXY_METHODS = frozenset([
        '__abs__',
        '__add__',
        '__and__',
        '__floordiv__',
        '__invert__',
        '__lshift__',
        '__mod__',
        '__mul__',
        '__neg__',
        '__or__',
        '__pos__',
        '__pow__',
        '__radd__',
        '__rand__',
        '__rfloordiv__',
        '__rlshift__',
        '__rmod__',
        '__rmul__',
        '__ror__',
        '__rpow__',
        '__rrshift__',
        '__rshift__',
        '__rsub__',
        '__rtruediv__',
        '__rxor__',
        '__sub__',
        '__truediv__',
        '__xor__',
    ])

    TUPLE_PROXY_METHODS = frozenset([
        '__divmod__',
        '__rdivmod__',
    ])

    for _proxy_name in SCALAR_PROXY_METHODS:
        locals()[_proxy_name] = _gen_scalar_proxy_method(_proxy_name)
    for _proxy_name in TUPLE_PROXY_METHODS:
        locals()[_proxy_name] = _gen_tuple_proxy_method(_proxy_name)
    del _proxy_name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({int(self)})'


class Int8(FixedWidthInt):
    BITS = 8
    SIGNED = True


class Int16(FixedWidthInt):
    BITS = 16
    SIGNED = True


class Int32(FixedWidthInt):
    BITS = 32
    SIGNED = True


class Int64(FixedWidthInt):
    BITS = 64
    SIGNED = True


class Int128(FixedWidthInt):
    BITS = 128
    SIGNED = True


class Uint8(FixedWidthInt):
    BITS = 8
    SIGNED = False


class Uint16(FixedWidthInt):
    BITS = 16
    SIGNED = False


class Uint32(FixedWidthInt):
    BITS = 32
    SIGNED = False


class Uint64(FixedWidthInt):
    BITS = 64
    SIGNED = False


class Uint128(FixedWidthInt):
    BITS = 128
    SIGNED = False



# type: ignore
# ruff: noqa
# flake8: noqa
# Generated from Json5.g4 by ANTLR 4.13.2
# encoding: utf-8
from omlish.antlr._runtime._all import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,15,74,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,1,0,3,0,16,8,0,1,0,1,0,1,1,1,1,1,1,1,1,5,1,24,8,1,10,1,12,1,27,
        9,1,1,1,3,1,30,8,1,1,1,1,1,1,1,1,1,3,1,36,8,1,1,2,1,2,1,2,1,2,1,
        3,1,3,1,4,1,4,1,4,1,4,1,4,3,4,49,8,4,1,5,1,5,1,5,1,5,5,5,55,8,5,
        10,5,12,5,58,9,5,1,5,3,5,61,8,5,1,5,1,5,1,5,1,5,3,5,67,8,5,1,6,3,
        6,70,8,6,1,6,1,6,1,6,0,0,7,0,2,4,6,8,10,12,0,2,3,0,9,10,12,12,14,
        14,1,0,11,12,78,0,15,1,0,0,0,2,35,1,0,0,0,4,37,1,0,0,0,6,41,1,0,
        0,0,8,48,1,0,0,0,10,66,1,0,0,0,12,69,1,0,0,0,14,16,3,8,4,0,15,14,
        1,0,0,0,15,16,1,0,0,0,16,17,1,0,0,0,17,18,5,0,0,1,18,1,1,0,0,0,19,
        20,5,1,0,0,20,25,3,4,2,0,21,22,5,2,0,0,22,24,3,4,2,0,23,21,1,0,0,
        0,24,27,1,0,0,0,25,23,1,0,0,0,25,26,1,0,0,0,26,29,1,0,0,0,27,25,
        1,0,0,0,28,30,5,2,0,0,29,28,1,0,0,0,29,30,1,0,0,0,30,31,1,0,0,0,
        31,32,5,3,0,0,32,36,1,0,0,0,33,34,5,1,0,0,34,36,5,3,0,0,35,19,1,
        0,0,0,35,33,1,0,0,0,36,3,1,0,0,0,37,38,3,6,3,0,38,39,5,4,0,0,39,
        40,3,8,4,0,40,5,1,0,0,0,41,42,7,0,0,0,42,7,1,0,0,0,43,49,5,10,0,
        0,44,49,3,12,6,0,45,49,3,2,1,0,46,49,3,10,5,0,47,49,5,9,0,0,48,43,
        1,0,0,0,48,44,1,0,0,0,48,45,1,0,0,0,48,46,1,0,0,0,48,47,1,0,0,0,
        49,9,1,0,0,0,50,51,5,5,0,0,51,56,3,8,4,0,52,53,5,2,0,0,53,55,3,8,
        4,0,54,52,1,0,0,0,55,58,1,0,0,0,56,54,1,0,0,0,56,57,1,0,0,0,57,60,
        1,0,0,0,58,56,1,0,0,0,59,61,5,2,0,0,60,59,1,0,0,0,60,61,1,0,0,0,
        61,62,1,0,0,0,62,63,5,6,0,0,63,67,1,0,0,0,64,65,5,5,0,0,65,67,5,
        6,0,0,66,50,1,0,0,0,66,64,1,0,0,0,67,11,1,0,0,0,68,70,5,13,0,0,69,
        68,1,0,0,0,69,70,1,0,0,0,70,71,1,0,0,0,71,72,7,1,0,0,72,13,1,0,0,
        0,9,15,25,29,35,48,56,60,66,69
    ]

class Json5Parser ( Parser ):

    grammarFileName = "Json5.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'{'", "','", "'}'", "':'", "'['", "']'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "SINGLE_LINE_COMMENT", 
                      "MULTI_LINE_COMMENT", "LITERAL", "STRING", "NUMBER", 
                      "NUMERIC_LITERAL", "SYMBOL", "IDENTIFIER", "WS" ]

    RULE_json5 = 0
    RULE_obj = 1
    RULE_pair = 2
    RULE_key = 3
    RULE_value = 4
    RULE_arr = 5
    RULE_number = 6

    ruleNames =  [ "json5", "obj", "pair", "key", "value", "arr", "number" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    SINGLE_LINE_COMMENT=7
    MULTI_LINE_COMMENT=8
    LITERAL=9
    STRING=10
    NUMBER=11
    NUMERIC_LITERAL=12
    SYMBOL=13
    IDENTIFIER=14
    WS=15

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Json5Context(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(Json5Parser.EOF, 0)

        def value(self):
            return self.getTypedRuleContext(Json5Parser.ValueContext,0)


        def getRuleIndex(self):
            return Json5Parser.RULE_json5

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterJson5" ):
                listener.enterJson5(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitJson5" ):
                listener.exitJson5(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitJson5" ):
                return visitor.visitJson5(self)
            else:
                return visitor.visitChildren(self)




    def json5(self):

        localctx = Json5Parser.Json5Context(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_json5)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 15
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 15906) != 0):
                self.state = 14
                self.value()


            self.state = 17
            self.match(Json5Parser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ObjContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def pair(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(Json5Parser.PairContext)
            else:
                return self.getTypedRuleContext(Json5Parser.PairContext,i)


        def getRuleIndex(self):
            return Json5Parser.RULE_obj

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObj" ):
                listener.enterObj(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObj" ):
                listener.exitObj(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObj" ):
                return visitor.visitObj(self)
            else:
                return visitor.visitChildren(self)




    def obj(self):

        localctx = Json5Parser.ObjContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_obj)
        self._la = 0 # Token type
        try:
            self.state = 35
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,3,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 19
                self.match(Json5Parser.T__0)
                self.state = 20
                self.pair()
                self.state = 25
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,1,self._ctx)
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt==1:
                        self.state = 21
                        self.match(Json5Parser.T__1)
                        self.state = 22
                        self.pair() 
                    self.state = 27
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,1,self._ctx)

                self.state = 29
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==2:
                    self.state = 28
                    self.match(Json5Parser.T__1)


                self.state = 31
                self.match(Json5Parser.T__2)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 33
                self.match(Json5Parser.T__0)
                self.state = 34
                self.match(Json5Parser.T__2)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PairContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def key(self):
            return self.getTypedRuleContext(Json5Parser.KeyContext,0)


        def value(self):
            return self.getTypedRuleContext(Json5Parser.ValueContext,0)


        def getRuleIndex(self):
            return Json5Parser.RULE_pair

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPair" ):
                listener.enterPair(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPair" ):
                listener.exitPair(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPair" ):
                return visitor.visitPair(self)
            else:
                return visitor.visitChildren(self)




    def pair(self):

        localctx = Json5Parser.PairContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_pair)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 37
            self.key()
            self.state = 38
            self.match(Json5Parser.T__3)
            self.state = 39
            self.value()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class KeyContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(Json5Parser.STRING, 0)

        def IDENTIFIER(self):
            return self.getToken(Json5Parser.IDENTIFIER, 0)

        def LITERAL(self):
            return self.getToken(Json5Parser.LITERAL, 0)

        def NUMERIC_LITERAL(self):
            return self.getToken(Json5Parser.NUMERIC_LITERAL, 0)

        def getRuleIndex(self):
            return Json5Parser.RULE_key

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterKey" ):
                listener.enterKey(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitKey" ):
                listener.exitKey(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitKey" ):
                return visitor.visitKey(self)
            else:
                return visitor.visitChildren(self)




    def key(self):

        localctx = Json5Parser.KeyContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_key)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 41
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 22016) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ValueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(Json5Parser.STRING, 0)

        def number(self):
            return self.getTypedRuleContext(Json5Parser.NumberContext,0)


        def obj(self):
            return self.getTypedRuleContext(Json5Parser.ObjContext,0)


        def arr(self):
            return self.getTypedRuleContext(Json5Parser.ArrContext,0)


        def LITERAL(self):
            return self.getToken(Json5Parser.LITERAL, 0)

        def getRuleIndex(self):
            return Json5Parser.RULE_value

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterValue" ):
                listener.enterValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitValue" ):
                listener.exitValue(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitValue" ):
                return visitor.visitValue(self)
            else:
                return visitor.visitChildren(self)




    def value(self):

        localctx = Json5Parser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_value)
        try:
            self.state = 48
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [10]:
                self.enterOuterAlt(localctx, 1)
                self.state = 43
                self.match(Json5Parser.STRING)
                pass
            elif token in [11, 12, 13]:
                self.enterOuterAlt(localctx, 2)
                self.state = 44
                self.number()
                pass
            elif token in [1]:
                self.enterOuterAlt(localctx, 3)
                self.state = 45
                self.obj()
                pass
            elif token in [5]:
                self.enterOuterAlt(localctx, 4)
                self.state = 46
                self.arr()
                pass
            elif token in [9]:
                self.enterOuterAlt(localctx, 5)
                self.state = 47
                self.match(Json5Parser.LITERAL)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ArrContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def value(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(Json5Parser.ValueContext)
            else:
                return self.getTypedRuleContext(Json5Parser.ValueContext,i)


        def getRuleIndex(self):
            return Json5Parser.RULE_arr

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterArr" ):
                listener.enterArr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitArr" ):
                listener.exitArr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitArr" ):
                return visitor.visitArr(self)
            else:
                return visitor.visitChildren(self)




    def arr(self):

        localctx = Json5Parser.ArrContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_arr)
        self._la = 0 # Token type
        try:
            self.state = 66
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,7,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 50
                self.match(Json5Parser.T__4)
                self.state = 51
                self.value()
                self.state = 56
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,5,self._ctx)
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt==1:
                        self.state = 52
                        self.match(Json5Parser.T__1)
                        self.state = 53
                        self.value() 
                    self.state = 58
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,5,self._ctx)

                self.state = 60
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==2:
                    self.state = 59
                    self.match(Json5Parser.T__1)


                self.state = 62
                self.match(Json5Parser.T__5)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 64
                self.match(Json5Parser.T__4)
                self.state = 65
                self.match(Json5Parser.T__5)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NumberContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NUMERIC_LITERAL(self):
            return self.getToken(Json5Parser.NUMERIC_LITERAL, 0)

        def NUMBER(self):
            return self.getToken(Json5Parser.NUMBER, 0)

        def SYMBOL(self):
            return self.getToken(Json5Parser.SYMBOL, 0)

        def getRuleIndex(self):
            return Json5Parser.RULE_number

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumber" ):
                listener.enterNumber(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumber" ):
                listener.exitNumber(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNumber" ):
                return visitor.visitNumber(self)
            else:
                return visitor.visitChildren(self)




    def number(self):

        localctx = Json5Parser.NumberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_number)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 69
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==13:
                self.state = 68
                self.match(Json5Parser.SYMBOL)


            self.state = 71
            _la = self._input.LA(1)
            if not(_la==11 or _la==12):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








# ruff: noqa: UP006 UP007
# @omlish-lite
import logging
import typing as ta


class ListHandler(logging.Handler):
    def __init__(self) -> None:
        super().__init__()
        self.records: ta.List[logging.LogRecord] = []

    def emit(self, record: logging.LogRecord) -> None:
        self.records.append(record)



"""
These attempt to punch through *all* wrappers. The the usecases involve having things like bound methods, classmethods,
functools.partial instances, and needing to get the metadata of the underlying thing the end-user wrote.

TODO:
 - re type targets:
  - unwrap instances of objects to their types?
  - merge mro?
  - are these better left up to callers? too usecase-specific to favor either way?
"""
import threading
import types
import typing as ta

from . import check
from . import lang


T = ta.TypeVar('T')


##


class ObjectMetadata(lang.Abstract):
    pass


class ObjectMetadataTarget(lang.Abstract):
    pass


##


_VALID_OBJECT_METADATA_TARGET_TYPES: tuple[type, ...] = (
    type,

    types.FunctionType,
    # *not* types.MethodType - must unwrap to unbound class func
    # *not* functools.partial - must unwrap to underlying func

    ObjectMetadataTarget,
)


class ObjectMetadataTargetTypeError(TypeError):
    pass


def _unwrap_object_metadata_target(obj: ta.Any) -> ta.Any:
    tgt: ta.Any = obj
    tgt = lang.unwrap_func(tgt)

    if not isinstance(tgt, _VALID_OBJECT_METADATA_TARGET_TYPES):
        raise ObjectMetadataTargetTypeError(tgt)

    return tgt


##


_OBJECT_METADATA_LOCK = threading.RLock()

_OBJECT_METADATA_ATTR = '__' + __name__.replace('.', '_') + '__metadata__'


def append_object_metadata(obj: T, *mds: ObjectMetadata) -> T:
    for md in mds:
        check.isinstance(md, ObjectMetadata)

    tgt = _unwrap_object_metadata_target(obj)
    dct = tgt.__dict__

    if isinstance(dct, types.MappingProxyType):
        try:
            lst = dct[_OBJECT_METADATA_ATTR]
        except KeyError:
            with _OBJECT_METADATA_LOCK:
                try:
                    lst = dct[_OBJECT_METADATA_ATTR]
                except KeyError:
                    setattr(tgt, _OBJECT_METADATA_ATTR, lst := [])

    else:
        lst = dct.setdefault(_OBJECT_METADATA_ATTR, [])

    lst.extend(mds)
    return obj


def get_object_metadata(obj: ta.Any, *, strict: bool = False) -> ta.Sequence[ObjectMetadata]:
    try:
        tgt = _unwrap_object_metadata_target(obj)
    except ObjectMetadataTargetTypeError:
        if not strict:
            return ()
        raise

    try:
        dct = tgt.__dict__
    except AttributeError:
        return ()

    return dct.get(_OBJECT_METADATA_ATTR, ())


##


class DecoratorObjectMetadata(ObjectMetadata, lang.Abstract):
    _OBJECT_METADATA_TARGET_TYPES: ta.ClassVar[tuple[type, ...] | None] = None

    def __init_subclass__(
            cls,
            *,
            object_metadata_target_types: ta.Iterable[type] | None = None,
            **kwargs: ta.Any,
    ) -> None:
        super().__init_subclass__(**kwargs)

        if object_metadata_target_types is not None:
            tts = tuple(object_metadata_target_types)
            for tt in tts:
                check.issubclass(tt, _VALID_OBJECT_METADATA_TARGET_TYPES)
            setattr(cls, '_OBJECT_METADATA_TARGET_TYPES', tts)

    def __call__(self, obj: T) -> T:
        tgt: ta.Any = obj
        if (tts := type(self)._OBJECT_METADATA_TARGET_TYPES) is not None:  # noqa
            tgt = _unwrap_object_metadata_target(tgt)
            check.isinstance(tgt, tts)

        append_object_metadata(tgt, self)
        return obj


#


class ClassDecoratorObjectMetadata(
    DecoratorObjectMetadata,
    lang.Abstract,
    object_metadata_target_types=[type],
):
    pass


class FunctionDecoratorObjectMetadata(
    DecoratorObjectMetadata,
    lang.Abstract,
    object_metadata_target_types=[types.FunctionType],
):
    pass



"""
TODO:
 - !! reconcile A().f() with A.f(A())
  - unbound descriptor *should* still hit instance cache
 - integrate / expose with collections.cache
 - weakrefs (selectable by arg)
 - more rigorous descriptor pickling
  - must support free functions (which have no instance nor owner)
  - 'staticmethod' or effective equiv - which must resolve to the shared instance
   - and must be transient?
 - use __transient_dict__ to support common state nuking
"""
import dataclasses as dc
import functools
import inspect
import typing as ta

from ..classes.abstract import Abstract
from ..contextmanagers import DefaultLockable
from ..contextmanagers import default_lock
from ..descriptors import unwrap_func
from ..descriptors import unwrap_func_with_partials
from ..params import KwargsParam
from ..params import Param
from ..params import ParamSeparator
from ..params import ParamSpec
from ..params import ValParam
from ..params import param_render


P = ta.ParamSpec('P')
T = ta.TypeVar('T')
CallableT = ta.TypeVar('CallableT', bound=ta.Callable)
CacheKeyMaker: ta.TypeAlias = ta.Callable[..., tuple]


##


def _nullary_cache_key_maker():
    return ()


def _self_cache_key_maker(self):
    return (self,)


_PRE_MADE_CACHE_KEY_MAKERS = [
    _nullary_cache_key_maker,
    _self_cache_key_maker,
]


_PRE_MADE_CACHE_KEY_MAKERS_BY_PARAM_SPEC: ta.Mapping[ParamSpec, CacheKeyMaker] = {
    ParamSpec.inspect(fn): fn  # type: ignore
    for fn in _PRE_MADE_CACHE_KEY_MAKERS
}


##


def _make_unwrapped_cache_key_maker(
        fn: ta.Callable,
        *,
        bound: bool = False,
) -> CacheKeyMaker:
    if inspect.isgeneratorfunction(fn) or inspect.iscoroutinefunction(fn):
        raise TypeError(fn)

    ps = ParamSpec.inspect(
        fn,
        offset=1 if bound else 0,
        strip_annotations=True,
    )

    if not len(ps):
        return _nullary_cache_key_maker

    if not ps.has_defaults:
        try:
            return _PRE_MADE_CACHE_KEY_MAKERS_BY_PARAM_SPEC[ps]
        except KeyError:
            pass

    builtin_pfx = '__cache_key_maker__'
    ns = {
        (builtin_tuple := builtin_pfx + 'tuple'): tuple,
        (builtin_sorted := builtin_pfx + 'sorted'): sorted,
    }

    src_params = []
    src_vals = []
    kwargs_name = None
    for p in ps:
        if isinstance(p, ParamSeparator):
            src_params.append(p.value)
            continue

        if not isinstance(p, Param):
            raise TypeError(p)

        if isinstance(p, ValParam) and p.default.present:
            ns[p.name] = p.default

        src_params.append(param_render(
            p,
            render_default=lambda _: p.name,  # noqa
        ))

        if isinstance(p, KwargsParam):
            kwargs_name = p.name
        else:
            src_vals.append(p.name)

    if kwargs_name is not None:
        src_vals.append(f'{builtin_tuple}({builtin_sorted}({kwargs_name}.items()))')

    rendered = (
        f'def __func__({", ".join(src_params)}):\n'
        f'    return ({", ".join(src_vals)}{"," if len(src_vals) == 1 else ""})\n'
    )
    exec(rendered, ns)

    kfn: CacheKeyMaker = ns['__func__']  # type: ignore[assignment]
    return kfn


def _make_cache_key_maker(
        fn: ta.Any,
        *,
        bound: bool = False,
):
    fn, partials = unwrap_func_with_partials(fn)

    kfn = _make_unwrapped_cache_key_maker(fn, bound=bound)

    for part in partials[::-1]:
        kfn = functools.partial(kfn, *part.args, **part.keywords)

    return kfn


##


class _CachedFunction(ta.Generic[T], Abstract):
    @dc.dataclass(frozen=True)
    class Opts:
        map_maker: ta.Callable[[], ta.MutableMapping] = dict
        lock: DefaultLockable = None
        transient: bool = False

    def __init__(
            self,
            fn: ta.Callable[P, T],
            *,
            opts: Opts = Opts(),
            key_maker: ta.Callable[..., tuple] | None = None,
            values: ta.MutableMapping | None = None,
            value_fn: ta.Callable[P, T] | None = None,
    ) -> None:
        super().__init__()

        self._fn = (fn,)
        self._opts = opts
        self._key_maker = key_maker if key_maker is not None else _make_cache_key_maker(fn)

        self._lock = default_lock(opts.lock, False)() if opts.lock is not None else None
        self._values = values if values is not None else opts.map_maker()
        self._value_fn = value_fn if value_fn is not None else fn
        functools.update_wrapper(self, fn)

    @property
    def _fn(self):
        return self.__fn

    @_fn.setter
    def _fn(self, x):
        self.__fn = x

    def reset(self) -> None:
        self._values = {}

    def __bool__(self) -> bool:
        raise TypeError

    def __call__(self, *args, **kwargs) -> T:
        k = self._key_maker(*args, **kwargs)

        try:
            return self._values[k]
        except KeyError:
            pass

        if self._lock is not None:
            with self._lock:
                try:
                    return self._values[k]
                except KeyError:
                    pass

                value = self._value_fn(*args, **kwargs)

        else:
            value = self._value_fn(*args, **kwargs)

        self._values[k] = value
        return value


#


class _FreeCachedFunction(_CachedFunction[T]):
    @classmethod
    def _unpickle(
            cls,
            fn,
            opts,
            values,
    ):
        return cls(
            fn,
            opts=opts,
            values=values,
        )

    def __reduce__(self):
        return (
            _FreeCachedFunction._unpickle,
            (
                self._fn,
                self._opts,
                self._values if not self._opts.transient else None,
            ),
        )


#


class _DescriptorCachedFunction(_CachedFunction[T]):
    def __init__(
            self,
            fn: ta.Callable[P, T],
            scope: ta.Any,  # classmethod | None
            *,
            instance: ta.Any = None,
            owner: ta.Any = None,
            name: str | None = None,
            **kwargs,
    ) -> None:
        super().__init__(fn, **kwargs)

        self._scope = scope
        self._instance = instance
        self._owner = owner
        self._name = name if name is not None else unwrap_func(fn).__name__
        self._bound_key_maker = None

    @classmethod
    def _unpickle(
            cls,
            scope,
            instance,
            owner,
            name,
            values,
    ):
        if scope is not None:
            raise NotImplementedError

        if instance is None:
            raise RuntimeError
        obj = type(instance)

        desc: _DescriptorCachedFunction = object.__getattribute__(obj, name)
        if not isinstance(desc, cls):
            raise TypeError(desc)
        if (desc._instance is not None or desc._owner is not None):
            raise RuntimeError

        return desc._bind(
            instance,
            owner,
            values=values,
        )

    def __reduce__(self):
        if self._scope is not None:
            raise NotImplementedError

        if not (self._instance is not None or self._owner is not None):
            raise RuntimeError

        return (
            _DescriptorCachedFunction._unpickle,
            (
                self._scope,
                self._instance,
                self._owner,
                self._name,
                self._values if not self._opts.transient else None,
            ),
        )

    def _bind(
            self,
            instance,
            owner=None,
            *,
            values: ta.MutableMapping | None = None,
    ):
        scope = self._scope
        if owner is self._owner and (instance is self._instance or scope is classmethod):
            return self

        fn, = self._fn
        name = self._name
        bound_fn = fn.__get__(instance, owner)
        if self._bound_key_maker is None:
            self._bound_key_maker = _make_cache_key_maker(fn, bound=True)

        bound = self.__class__(
            fn,
            scope,
            opts=self._opts,
            instance=instance,
            owner=owner,
            name=name,
            key_maker=self._bound_key_maker,
            # values=None if scope is classmethod else self._values,  # FIXME: ?
            values=values,
            value_fn=bound_fn,
        )

        if scope is classmethod and owner is not None:
            setattr(owner, name, bound)
        elif instance is not None:
            instance.__dict__[name] = bound

        return bound

    def __get__(self, instance, owner=None):
        return self._bind(instance, owner)


#


def cached_function(fn=None, **kwargs):  # noqa
    if fn is None:
        return functools.partial(cached_function, **kwargs)
    opts = _CachedFunction.Opts(**kwargs)
    if isinstance(fn, staticmethod):
        return _FreeCachedFunction(fn, opts=opts, value_fn=unwrap_func(fn))
    scope = classmethod if isinstance(fn, classmethod) else None
    return _DescriptorCachedFunction(fn, scope, opts=opts)


def static_init(fn: CallableT) -> CallableT:
    fn = cached_function(fn)
    fn()
    return fn



class Error(Exception):
    pass


class LineEmptyError(Error):
    pass


class BadCharactersError(Error):
    pass


class TagsTooLongError(Error):
    pass


class InvalidTagContentError(Error):
    pass


class CommandMissingError(Error):
    pass


class BadParamError(Error):
    pass


class MalformedNuhError(Error):
    pass



# ruff: noqa: UP006 UP007
# @omlish-lite
import abc
import dataclasses as dc
import http.server
import logging
import typing as ta

from ..sockets.addresses import SocketAddress
from .parsing import HttpHeaders


HttpHandler = ta.Callable[['HttpHandlerRequest'], 'HttpHandlerResponse']  # ta.TypeAlias
HttpHandlerResponseData = ta.Union[bytes, 'HttpHandlerResponseStreamedData']  # ta.TypeAlias  # noqa


##


@dc.dataclass(frozen=True)
class HttpHandlerRequest:
    client_address: SocketAddress
    method: str
    path: str
    headers: HttpHeaders
    data: ta.Optional[bytes]


@dc.dataclass(frozen=True)
class HttpHandlerResponse:
    status: ta.Union[http.HTTPStatus, int]

    headers: ta.Optional[ta.Mapping[str, str]] = None
    data: ta.Optional[HttpHandlerResponseData] = None
    close_connection: ta.Optional[bool] = None

    def close(self) -> None:
        if isinstance(d := self.data, HttpHandlerResponseStreamedData):
            d.close()


@dc.dataclass(frozen=True)
class HttpHandlerResponseStreamedData:
    iter: ta.Iterable[bytes]
    length: ta.Optional[int] = None

    def close(self) -> None:
        if hasattr(d := self.iter, 'close'):
            d.close()  # noqa


class HttpHandlerError(Exception):
    pass


class UnsupportedMethodHttpHandlerError(Exception):
    pass


class HttpHandler_(abc.ABC):  # noqa
    @abc.abstractmethod
    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        raise NotImplementedError


##


@dc.dataclass(frozen=True)
class LoggingHttpHandler(HttpHandler_):
    handler: HttpHandler
    log: logging.Logger
    level: int = logging.DEBUG

    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        self.log.log(self.level, '%r', req)
        resp = self.handler(req)
        self.log.log(self.level, '%r', resp)
        return resp


@dc.dataclass(frozen=True)
class ExceptionLoggingHttpHandler(HttpHandler_):
    handler: HttpHandler
    log: logging.Logger
    message: ta.Union[str, ta.Callable[[HttpHandlerRequest, BaseException], str]] = 'Error in http handler'

    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        try:
            return self.handler(req)
        except Exception as e:  # noqa
            if callable(msg := self.message):
                msg = msg(req, e)
            self.log.exception(msg)
            raise


##


@dc.dataclass(frozen=True)
class BytesResponseHttpHandler(HttpHandler_):
    data: bytes

    status: ta.Union[http.HTTPStatus, int] = 200
    content_type: ta.Optional[str] = 'application/octet-stream'
    headers: ta.Optional[ta.Mapping[str, str]] = None
    close_connection: bool = True

    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        return HttpHandlerResponse(
            status=self.status,
            headers={
                **({'Content-Type': self.content_type} if self.content_type else {}),
                'Content-Length': str(len(self.data)),
                **(self.headers or {}),
            },
            data=self.data,
            close_connection=self.close_connection,
        )


@dc.dataclass(frozen=True)
class StringResponseHttpHandler(HttpHandler_):
    data: str

    status: ta.Union[http.HTTPStatus, int] = 200
    content_type: ta.Optional[str] = 'text/plain; charset=utf-8'
    headers: ta.Optional[ta.Mapping[str, str]] = None
    close_connection: bool = True

    def __call__(self, req: HttpHandlerRequest) -> HttpHandlerResponse:
        data = self.data.encode('utf-8')
        return HttpHandlerResponse(
            status=self.status,
            headers={
                **({'Content-Type': self.content_type} if self.content_type else {}),
                'Content-Length': str(len(data)),
                **(self.headers or {}),
            },
            data=data,
            close_connection=self.close_connection,
        )



import typing as ta

from .. import dataclasses as dc
from .. import lang


LifecycleT = ta.TypeVar('LifecycleT', bound='Lifecycle')
LifecycleCallback: ta.TypeAlias = ta.Callable[[LifecycleT], None]


class Lifecycle:
    def lifecycle_construct(self) -> None:
        pass

    def lifecycle_start(self) -> None:
        pass

    def lifecycle_stop(self) -> None:
        pass

    def lifecycle_destroy(self) -> None:
        pass


@dc.dataclass(frozen=True, kw_only=True)
class CallbackLifecycle(Lifecycle, lang.Final, ta.Generic[LifecycleT]):
    on_construct: LifecycleCallback['CallbackLifecycle[LifecycleT]'] | None = None
    on_start: LifecycleCallback['CallbackLifecycle[LifecycleT]'] | None = None
    on_stop: LifecycleCallback['CallbackLifecycle[LifecycleT]'] | None = None
    on_destroy: LifecycleCallback['CallbackLifecycle[LifecycleT]'] | None = None

    @ta.override
    def lifecycle_construct(self) -> None:
        if self.on_construct is not None:
            self.on_construct(self)

    @ta.override
    def lifecycle_start(self) -> None:
        if self.on_start is not None:
            self.on_start(self)

    @ta.override
    def lifecycle_stop(self) -> None:
        if self.on_stop is not None:
            self.on_stop(self)

    @ta.override
    def lifecycle_destroy(self) -> None:
        if self.on_destroy is not None:
            self.on_destroy(self)



# ruff: noqa: UP006 UP007
# @omlish-lite
import contextlib
import io
import typing as ta

from ..lite.check import check


class IndentWriter:
    DEFAULT_INDENT = ' ' * 4

    def __init__(
            self,
            *,
            buf: ta.Optional[io.StringIO] = None,
            indent: ta.Optional[str] = None,
    ) -> None:
        super().__init__()

        self._buf = buf if buf is not None else io.StringIO()
        self._indent = check.isinstance(indent, str) if indent is not None else self.DEFAULT_INDENT
        self._level = 0
        self._has_indented = False

    @contextlib.contextmanager
    def indent(self, num: int = 1) -> ta.Iterator[None]:
        self._level += num
        try:
            yield
        finally:
            self._level -= num

    def write(self, s: str) -> None:
        indent = self._indent * self._level
        i = 0
        while i < len(s):
            if not self._has_indented:
                self._buf.write(indent)
                self._has_indented = True
            try:
                n = s.index('\n', i)
            except ValueError:
                self._buf.write(s[i:])
                break
            self._buf.write(s[i:n + 1])
            self._has_indented = False
            i = n + 2

    def getvalue(self) -> str:
        return self._buf.getvalue()
